angular.module('beamng.apps')
  .directive('dashcamui', [function () {
    return {
      templateUrl: '/ui/modules/apps/dashcamui/app.html',
      replace: true,
      restrict: 'EA',
      scope: true,
      link: function (scope, element, attrs) {
        //streams
        var streamsList = ['electrics', 'sensors'];
        StreamsManager.add(streamsList);
        //config values
        let gpsDelay = 2
        let ampmclock = 0
        let showSpeed = 1
        let showCoords = 1
        let showGForce = 0
        let orangeColor = 0
        let fontSize = 3
        let fontName = 'Roboto'

        //values
        let date
        let hours = -1
        let minutes = -1
        let seconds = -1

        let gpsSpeed = 0
        let count = 0
        let sum = 0

        let longitude
        let latitude

        let maxGForce = 0 //max GForce recorded in beetween overlay updates
        let gForce = 0 //current value displayed

        element.ready(function () { init() });
        function init() {
          //get local date
          getLocalDate()

          //get in-game time for clock
          getGameTime()

          //random coordinates to start with
          latitude = (Math.random() * 180) - 90;
          longitude = (Math.random() * 360) - 180;

          //click event for showing options
          document.getElementById("btn").addEventListener("click", function () {
            document.getElementById("options").style.display = document.getElementById("options").style.display == "block" ? "none" : "block";
          });
          //click event for applying options
          document.getElementById("btnApply").addEventListener("click", function () {
            applyOptions(); saveUserPref();
          });
          document.getElementById("resetDate").onclick = getLocalDate

          //reading and applying config
          readUserPref()
          applyConfig()
        }

        //getting values from options window, then calling applyConfig()
        function applyOptions() {
          var options = document.getElementById('optionsForm').elements;
          for (var i = 0; i < options.length; i++) {
            switch (i) {
              case 0:
                date = options[i].value;
                break;
              case 1:
                gpsDelay = options[i].value;
                break;
              case 2:
                fontSize = options[i].value
                break;
              case 3:
                showSpeed = options[i].checked;
                break;
              case 4:
                showCoords = options[i].checked;
                break;
              case 5:
                showGForce = options[i].checked;
                break;
              case 6:
                ampmclock = options[i].checked;
                break;
              case 7:
                orangeColor = options[i].checked;
                break;
            }
          }
          applyConfig()
          document.getElementById("options").style.display = "none";
        }
        //updating actual elements
        function applyConfig() {
          document.getElementById("overlay").style.fontFamily = fontName;
          if (fontSize < 10) { fontSize = 10 }
          document.getElementById("overlay").style.fontSize = fontSize / 10 + "em";
          let shadow = fontSize / 10 - 1
          document.getElementById("overlay").style.textShadow = `${-shadow}px ${-shadow}px 0 #000, ${shadow}px ${-shadow}px 0 #000, ${-shadow}px ${shadow}px 0 #000, ${shadow}px ${shadow}px 0 #000`;
          document.getElementById("overlay").style.color = orangeColor ? "#ff6600" : "#ffffff"
          document.getElementById("logo").src = orangeColor ? "/ui/modules/apps/dashcamui/img/BeamNG-logo-icon-2016.svg" : "/ui/modules/apps/dashcamui/img/BeamNG-logo-icon-2016-white.svg"
        }
        function readUserPref() {
          let value = localStorage.getItem('dashcamUI-options');
          if (value) {
            let setting = JSON.parse(value);

            var options = document.getElementById('optionsForm').elements;
            for (var i = 0; i < options.length; i++) {
              switch (i) {
                case 0:
                  options[i].value = setting.date;
                  break;
                case 1:
                  options[i].value = setting.gpsDelay;
                  break;
                case 2:
                  options[i].value = setting.fontSize;
                  break;
                case 3:
                  options[i].checked = setting.showSpeed;
                  break;
                case 4:
                  options[i].checked = setting.showCoords;
                  break;
                case 5:
                  options[i].checked = setting.showGForce;
                  break;
                case 6:
                  options[i].checked = setting.ampmclock;
                  break;
                case 7:
                  options[i].checked = setting.orangeColor;
                  break;
              }
            }
          }
          else //no user config detected, using default settings
          {
            document.getElementById("enterDate").value = date;
            document.getElementById("GPSdelay").value = 2;
            document.getElementById("fontSize").value = 30;
            document.getElementById("coords").checked = true;
            document.getElementById("speed").checked = true;
          }
          applyOptions()
        }
        function saveUserPref() {
          var options = {
            date: date,
            gpsDelay: gpsDelay,
            fontSize: fontSize,
            showSpeed: showSpeed,
            showCoords: showCoords,
            showGForce: showGForce,
            ampmclock: ampmclock,
            orangeColor: orangeColor
            //fontName: fontName
          };
          localStorage.setItem('dashcamUI-options', JSON.stringify(options));
        }

        function getLocalDate() {
          let currentDate = new Date();
          let year = currentDate.getFullYear();
          let month = String(currentDate.getMonth() + 1).padStart(2, '0');
          let day = String(currentDate.getDate()).padStart(2, '0');
          date = `${year}-${month}-${day}`
          document.getElementById("enterDate").value = date;
        }
        function getGameTime() {
          bngApi.engineLua('core_environment.getTimeOfDay().time', function (data) {
            //data - actual time between ''0'' and ''1'' where ''0'' and ''1'' are noon (12pm) and 0.5 is midnight (12am)
            data = (12 + data * 24) % 24
            hours = Math.floor(data)
            minutes = Math.floor((data - hours) * 60)
          });
        }
        function convertTo12H(hour, minute) {
          let period = hour >= 12 ? ' PM' : ' AM'
          hour = hour % 12;
          hour = hour ? hour : 12;

          hour = String(hour).padStart(2, '0')
          minute = String(minute).padStart(2, '0')
          seconds = String(seconds).padStart(2, '0')

          return hour + ':' + minute + ':' + seconds + ' ' + period;
        }
        function calculateGForce(streams) {
          var gravity = streams.sensors.gravity
          gravity = gravity >= 0 ? Math.max(0.1, gravity) : Math.min(-0.1, gravity)
          let xg = Math.abs(streams.sensors.gx2) / -gravity;
          let yg = Math.abs(streams.sensors.gy2) / -gravity;
          let result = (xg + yg);
          //logic for getting the highest reading beetween overlay updates
          if (maxGForce < result) { maxGForce = result }
        }
        function calculateNewCoordinates(startLat, startLon, speed, directionDegrees, intervalMillis) {
          // Radius of the Earth in meters
          const R = 6378137;

          // Convert direction from degrees to radians
          const directionRadians = directionDegrees * Math.PI / 180;

          // Calculate the distance traveled in the given time interval
          const distance = speed / 10 * intervalMillis;

          // Calculate the new latitude using the Haversine formula
          const newLat = Math.asin(
            Math.sin(startLat) * Math.cos(distance / R) +
            Math.cos(startLat) * Math.sin(distance / R) * Math.cos(directionRadians)
          );

          // Calculate the new longitude using the Haversine formula
          const newLon = startLon + Math.atan2(
            Math.sin(directionRadians) * Math.sin(distance / R) * Math.cos(startLat),
            Math.cos(distance / R) - Math.sin(startLat) * Math.sin(newLat)
          );

          return [newLat, newLon];
        }
        function formatCoordinates(latitude, longitude) {
          // Convert latitude to degrees, minutes, seconds, and direction
          var latDegrees = Math.floor(Math.abs(latitude));
          var latMinutes = Math.floor((Math.abs(latitude) - latDegrees) * 60);
          var latSeconds = Math.round(((Math.abs(latitude) - latDegrees) * 3600) % 60);
          var latDirection = latitude >= 0 ? "N" : "S";

          // Convert longitude
          var lonDegrees = Math.floor(Math.abs(longitude));
          var lonMinutes = Math.floor((Math.abs(longitude) - lonDegrees) * 60);
          var lonSeconds = Math.round(((Math.abs(longitude) - lonDegrees) * 3600) % 60);
          var lonDirection = longitude >= 0 ? "E" : "W";

          // Format the coordinates string and return ready-to-print text
          return (String(latDegrees).padStart(3, '0') + "°" +
            String(latMinutes).padStart(3, '0') + "'" +
            String(latSeconds).padStart(3, '0') + "\"" +
            latDirection + " " +
            String(lonDegrees).padStart(3, '0') + "°" +
            String(lonMinutes).padStart(3, '0') + "'" +
            String(lonSeconds).padStart(3, '0') + "\"" +
            lonDirection);
        }

        function checkForUpdate() {
          let currentDate = new Date();
          if (currentDate.getSeconds() != seconds) {
            seconds = currentDate.getSeconds();
            if (seconds === 0) {
              minutes++
              if (minutes === 60) {
                minutes = 0;
                hours++;
                if (hours >= 24) {
                  hours = 0;
                }
              }
            }
            if (gpsDelay <= 1) { return true }
            return seconds % gpsDelay === 0
          }
          return gpsDelay === "0"
        }

        function getData(streams) {
          if (streams === undefined) return { val: 0, unit: '' }
          sum += streams.electrics.airspeed
          count++

          if (checkForUpdate()) {
            gpsSpeed = sum / count; //calculating avg speed for gps delay
            [latitude, longitude] = calculateNewCoordinates(latitude, longitude, gpsSpeed, streams.sensors.yaw * 180 / Math.PI + 180, Math.max(1000 / 30, gpsDelay * 1000))

            sum = count = 0;
            gForce = maxGForce; maxGForce = 0
          }

          let res = UiUnits.speed(gpsSpeed)

          res.val = showSpeed ? String(res.val.toFixed()).padStart(3, '0') + res.unit : ""

          res.coords = showCoords ? formatCoordinates(latitude, longitude) : ""

          res.date = date
          res.time = ampmclock ? res.time = convertTo12H(hours, minutes) : String(hours).padStart(2, '0') + ":" + String(minutes).padStart(2, '0') + ":" + String(seconds).padStart(2, '0')

          if (showGForce) { calculateGForce(streams); res.gforce = gForce.toFixed(1) + "G" }
          else { res.gforce = "" }

          return res
        }
        //clean up after closing the app
        scope.$on('$destroy', function () {
          StreamsManager.remove(streamsList);
          scope.stop;
        });
        //updating
        scope.$on('streamsUpdate', function (event, data) {
          scope.set = getData(data)
          scope.$apply()
        })
      }
    }
  }])